"""
navdoon.collector
-----------------
Define collectors, that collect Statsd requests and queue them to be
processed by the processor.
"""

import os
import socket
from abc import abstractmethod, ABCMeta
from threading import Event
from navdoon.pystdlib.queue import Queue
from navdoon.utils import LoggerMixIn


DEFAULT_PORT = 8125


class AbstractCollector(object):
    """Abstract base class for collectors"""

    __metaclass__ = ABCMeta

    def __init__(self):
        self._queue = Queue()

    @abstractmethod
    def start(self):
        raise NotImplementedError

    @abstractmethod
    def wait_until_queuing_requests(self, timeout=None):
        raise NotImplementedError

    @abstractmethod
    def shutdown(self):
        raise NotImplementedError

    @abstractmethod
    def wait_until_shutdown(self, timeout=None):
        raise NotImplementedError

    @property
    def queue(self):
        return self._queue

    @queue.setter
    def queue(self, value):
        for method in ('put_nowait', ):
            if not callable(getattr(value, method, None)):
                raise ValueError(
                    "Invalid queue for collector. Queue is missing "
                    "method '{}'".format(method))
        self._queue = value


class SocketServer(LoggerMixIn, AbstractCollector):
    """Collect Statsd metrics via TCP/UDP socket"""

    default_port = DEFAULT_PORT

    def __init__(self, **kargs):
        AbstractCollector.__init__(self)
        LoggerMixIn.__init__(self)
        self.chunk_size = 65535
        self.socket_type = socket.SOCK_DGRAM
        self.socket_timeout = 1
        self.host = '127.0.0.1'
        self.port = self.__class__.default_port
        self.user = None
        self.group = None
        self.socket = None
        self._stop_queuing_requests = Event()
        self._queuing_requests = Event()
        self._shutdown = Event()
        self._should_shutdown = Event()
        self.configure(**kargs)
        self.log_signature = "collector.socket_server "

    def __del__(self):
        self._do_shutdown()

    def configure(self, **kargs):
        """Configure the server, setting attributes.
        Returns a list of attribute names that were affected
        """
        configured = []
        for key in ('host', 'port', 'user', 'group', 'socket_type'):
            if key in kargs:
                setattr(self, key, kargs[key])
                configured.append(key)
        return configured

    def start(self):
        self._bind_socket()
        try:
            while not self._should_shutdown.is_set():
                self._shutdown.clear()
                self._log("starting serving requests on {} {}:{}".format(
                    self.socket_type == socket.SOCK_STREAM and 'TCP' or 'UDP',
                    self.host, self.port))
                self._pre_start()
                if self.socket_type == socket.SOCK_STREAM:
                    self._queue_requests_tcp()
                else:
                    self._queue_requests_udp()
                self._post_start()
                self._log("stopped serving requests")
        finally:
            self._do_shutdown()

    def is_queuing_requests(self):
        return self._queuing_requests.is_set()

    def wait_until_queuing_requests(self, timeout=None):
        self._queuing_requests.wait(timeout)

    def shutdown(self):
        self._log_debug("shutting down ...")
        self._stop_queuing_requests.set()
        self._should_shutdown.set()

    def wait_until_shutdown(self, timeout=None):
        self._shutdown.wait(timeout)

    def _pre_start(self):
        pass

    def _queue_requests_udp(self):
        stop = self._stop_queuing_requests
        chunk_size = self.chunk_size
        sock = self.socket
        receive = sock.recv
        timeout_exception = socket.timeout
        enqueue = self._queue.put_nowait

        try:
            self._queuing_requests.set()
            while not stop.is_set():
                try:
                    data = receive(chunk_size)
                except timeout_exception:
                    data = None
                if data:
                    enqueue(data.decode())
        finally:
            self._queuing_requests.clear()

    def _queue_requests_tcp(self):
        stop = self._stop_queuing_requests
        buffer_size = self.chunk_size
        queue_put_nowait = self._queue.put_nowait
        shutdown_rdwr = socket.SHUT_RDWR
        socket_timeout_exception = socket.timeout

        def _enqueue_from_connection(conn):
            receive = conn.recv
            enqueue = queue_put_nowait
            socket_timeout = socket_timeout_exception
            try:
                while not stop.is_set():
                    try:
                        buff = receive(buffer_size)
                    except socket_timeout:
                        continue
                    if not buff:
                        break
                    enqueue(buff.decode())
            finally:
                conn.shutdown(shutdown_rdwr)
                conn.close()

        try:
            self._queuing_requests.set()
            while not stop.is_set():
                try:
                    connection = self.socket.accept()[0]
                except socket_timeout_exception:
                    continue
                _enqueue_from_connection(connection)
        finally:
            self._queuing_requests.clear()

    def _post_start(self):
        pass

    def _do_shutdown(self):
        self._close_socket()
        self._shutdown.set()

    def _bind_socket(self, sock=None):
        if not sock:
            sock = self._create_socket()
        self._close_socket()
        self._change_process_user_group()
        self.socket = sock
        self._log("bound to address {}:{}".format(*sock.getsockname()))

    def _create_socket(self):
        sock = socket.socket(socket.AF_INET, self.socket_type)
        sock.bind((self.host, self.port))
        sock.settimeout(self.socket_timeout)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        if self.socket_type == socket.SOCK_STREAM:
            sock.listen(5)
        return sock

    def _close_socket(self):
        sock = self.socket
        if sock:
            try:
                sock.shutdown(socket.SHUT_RDWR)
                sock.close()
            except socket.error:
                pass
        self.socket = None

    def _change_process_user_group(self):
        if self.user:
            self._log("changing process user to {}".format(self.user))
            os.seteuid(self.user)
        if self.group:
            self._log("changing process group to {}".format(self.group))
            os.setegid(self.group)



"""
navdoon.server
--------------
Define the Statsd server, that uses other components (collector, processor,
destination) to handle Statsd requests and flushe metrics to specified
destinations.
"""

import multiprocessing
from time import time, sleep
from threading import Thread, RLock, Event
from navdoon.pystdlib import queue
from navdoon.collector import AbstractCollector
from navdoon.utils import LoggerMixIn
from navdoon.processor import QueueProcessor


def validate_collectors(collectors):
    """Validate collectors to be usable by the server"""
    for collector in collectors:
        if not isinstance(collector, AbstractCollector):
            raise ValueError(
                "Collectors should extend AbstractCollector")


class Server(LoggerMixIn):
    """Statsd server"""

    def __init__(self):
        super(Server, self).__init__()
        self.log_signature = 'server '
        self._collectors = []
        self._running = Event()
        self._shutdown = Event()
        self._running_lock = RLock()
        self._shutdown_lock = RLock()
        self._queue = self._create_queue()
        self._queue_processor = None

    @property
    def queue_processor(self):
        return self._queue_processor

    @queue_processor.setter
    def queue_processor(self, value):
        if not isinstance(value, QueueProcessor):
            raise ValueError(
                "Invalid queue value. Processor should extend QueueProcessor")
        self._queue_processor = value

    def set_collectors(self, collectors):
        validate_collectors(collectors)
        self._collectors = collectors
        return self

    def start(self):
        if not self._collectors:
            raise Exception("Can not start Statsd server without a collector")
        if self._queue_processor is None:
            self._log_warn(
                "no queue processor provided. Creating a queue processor "
                "with no destinations")
            self._queue_processor = self.create_queue_processor()
        with self._running_lock:
            self._log("starting ...")
            self._share_queue()
            queue_thread = self._start_queue_processor()
            self._log_debug("started queue processor thread")
            try:
                self._start_collecting()
            finally:
                self._log("stopped, joining queue processor thread")
                queue_thread.join()
                self._running.clear()

    def _start_collecting(self):
        collector_threads = []
        self._log_debug("starting {} collectors ...".format(len(
            self._collectors)))
        for collector in self._collectors:
            thread = Thread(target=collector.start)
            collector_threads.append(thread)
            thread.start()
            collector.wait_until_queuing_requests()
        self._running.set()
        self._log("collectors are running")
        for thread in collector_threads:
            thread.join()

    def is_running(self):
        return self._running.is_set()

    def wait_until_running(self, timeout=None):
        self._running.wait(timeout)

    def shutdown(self, process_queue=True, timeout=None):
        with self._shutdown_lock:
            self._log("shutting down ...")
            start_time = time()
            self._shutdown_collectors(timeout)
            if self._queue_processor.is_processing():
                self._shutdown_queue_processor(process_queue, max(
                    0.1, timeout - (time() - start_time)) if timeout else None)
            self._close_queue()
            self._shutdown.set()

    def wait_until_shutdown(self, timeout=None):
        start = time()
        self._shutdown.wait(timeout)
        while True:
            if not self._running.is_set():
                break
            if timeout is not None and time() - start > timeout:
                raise Exception("Server shutdown timeout")
            sleep(0.5)

    def create_queue_processor(self):
        processor = QueueProcessor(self._queue)
        processor.logger = self.logger
        return processor

    def reload(self):
        raise NotImplementedError()

    @staticmethod
    def _use_multiprocessing():
        # FIXME: use multiprocessing if available
        # return available_cpus() > 1
        return False

    @classmethod
    def _create_queue(cls):
        return multiprocessing.Queue() if cls._use_multiprocessing(
        ) else queue.Queue()

    def _share_queue(self):
        queue_ = self._queue
        self._queue_processor.queue = queue_
        for collector in self._collectors:
            collector.queue = queue_

    def _close_queue(self):
        queue_ = self._queue
        if queue_:
            if callable(getattr(queue_, 'close', None)):
                queue_.close()
        self._queue = None

    def _start_queue_processor(self):
        if self._use_multiprocessing():
            self._log_debug("stating queue processor in a separate process")
            queue_process = multiprocessing.Process(
                target=self._queue_processor.process)
            queue_process.start()
        else:
            self._log_debug("stating queue processor in a thread")
            queue_process = Thread(target=self._queue_processor.process)
            queue_process.start()

        self._queue_processor.wait_until_processing(30)
        if not self._queue_processor.is_processing():
            self._log_error(
                "queue processor didn't start correctly time is up")
            self._queue_processor.shutdown()
            queue_process.join()
            raise Exception("Failed to start the queue processor")

        return queue_process

    def _shutdown_queue_processor(self, process=True, timeout=None):
        start_time = time()
        self._log_debug("shutting down the queue processor ...")
        self._queue.put_nowait(self._queue_processor.stop_process_token)
        if not process:
            self._queue_processor.shutdown()
        self._queue_processor.wait_until_shutdown(timeout)
        if self._queue_processor.is_processing():
            self._log_error(
                "Queue processor shutdown timeout after {} seconds".format(
                    time() - start_time))
            raise Exception(
                "Server shutdown timeout when shutting down processor")

    def _shutdown_collectors(self, timeout=None):
        start_time = time()
        if self._collectors:
            self._log_debug("shutting down {} collectors ...".format(len(
                self._collectors)))
            for collector in self._collectors:
                collector.shutdown()
                collector.wait_until_shutdown(timeout)
                if timeout is not None:
                    time_elapsed = time() - start_time
                    if time_elapsed > timeout:
                        self._log_error(
                            "Collectors shutdown timeout after "
                            "{} seconds".format(time_elapsed))
                        raise Exception(
                            "Server shutdown timed out when "
                            "shutting down collectors")
                    else:
                        timeout -= time_elapsed



"""
navdoon.destination.stream
--------------------------
Define destinations to flush metrics to streams
"""

import sys
from time import time
from navdoon.destination.abstract import AbstractDestination


class Stream(AbstractDestination):
    """Destination to flush metrics to stream"""

    def __init__(self, file_handle):
        self._file = None
        self.pattern = "{name} {value} {timestamp}"
        self.append = "\n"
        self.stream = file_handle

    @property
    def stream(self):
        return self._file

    @stream.setter
    def stream(self, file_):
        if not callable(getattr(file_, 'write', None)):
            raise ValueError("Destination stream should be file like object")
        self._file = file_

    def create_output_from_metrics(self, metrics):
        """Creates the output to be flushed, from the metrics"""
        requests = []
        for metric in metrics:
            name, value = metric[:2]
            timestamp = len(metric) > 2 and metric[2] or time()
            requests.append(self.pattern.format(name=name,
                                                value=value,
                                                timestamp=timestamp))
        return requests

    def flush(self, metrics):
        """Flush metrics to the stream"""
        lines = self.create_output_from_metrics(metrics)
        self._write_lines(lines)

    def _write_lines(self, lines):
        write = self._file.write
        append_ = self.append
        for line in lines:
            write(line + append_)
        self._file.flush()

    def __eq__(self, other):
        return self.stream == other.stream


class Stdout(Stream):
    """Destination to flush metrics to standard output"""

    def __init__(self):
        super(Stdout, self).__init__(sys.stdout)



"""
navdoon.destination
-------------------
Where the data received from Statsd will be
flushed to
"""
from navdoon.destination.abstract import AbstractDestination
from navdoon.destination.graphite import Graphite
from navdoon.destination.stream import Stream, Stdout



"""
navdoon.pystdlib.confirparser
-----------------------------
Abstract configparser/ConfigParser module from Python standard library
"""

try:
    from ConfigParser import *
except ImportError:
    from configparser import *



"""
navdoon.destination.abstract
----------------------------
Define an abstract base class for destinations
"""

from abc import abstractmethod, ABCMeta


class AbstractDestination(object):
    """Abstract base class for destinations"""

    __metaclass__ = ABCMeta

    @abstractmethod
    def flush(self, metrics):
        """Flush the metrics"""
        raise NotImplementedError



"""
navdoon.app
-----------
Define the application, what is executed and handles interations and signals.
The application combines all the other components into what the end user
can use as the final product.
"""

from __future__ import print_function

import socket
import sys
import logging
import logging.handlers
from argparse import ArgumentParser, FileType
from threading import RLock, Thread
from signal import signal, SIGINT, SIGTERM, SIGHUP
from time import sleep

import navdoon
from navdoon.pystdlib import configparser
from navdoon.server import Server
from navdoon.destination import Stdout, Graphite
from navdoon.collector import SocketServer, DEFAULT_PORT
from navdoon.utils import os_syslog_socket, LoggerMixIn

LOG_LEVEL_NAMES = ('DEBUG', 'INFO', 'WARN', 'ERROR', 'FATAL', 'CRITICAL')


def parse_config_file(file_):
    """Parse app configurations from contents of a file"""

    parser = configparser.SafeConfigParser()
    parser.readfp(file_)
    config = dict()
    for key, value in parser.items('navdoon'):
        if key == 'config':
            continue
        elif key == 'log-level' and value not in LOG_LEVEL_NAMES:
            raise ValueError(
                "Invalid log level '{}' in configuration file '{}'".format(
                    value, file_.name))
        config[key.replace('-', '_')] = value
    return config


def default_syslog_socket():
    return os_syslog_socket() or 'localhost:{}'.format(
                                logging.handlers.SYSLOG_UDP_PORT)


class App(LoggerMixIn):
    """Navdoon application, configure and use the components based on the
    provided arguments and configurations.
    """

    def __init__(self, args):
        super(App, self).__init__()
        self.log_signature = "navdoon.app "
        self._config = dict()
        self._args = args
        self._server = None
        self._run_lock = RLock()
        self._shutdown_lock = RLock()
        self._reload_lock = RLock()
        self._configure(args)

    def __del__(self):
        self.shutdown(1)
        self._close_logger()

    @staticmethod
    def get_description():
        return "{} v{}\n{}".format(navdoon.__title__, navdoon.__version__,
                                   navdoon.__summary__)

    @staticmethod
    def get_default_config():
        return dict(config=None,
                    log_level='INFO',
                    log_file=None,
                    log_stderr=False,
                    log_syslog=False,
                    syslog_socket= default_syslog_socket(),
                    flush_interval=1,
                    flush_stdout=False,
                    flush_graphite='',
                    collect_udp='',
                    collect_tcp='')

    def get_args(self):
        return self._args

    def get_config(self):
        return self._config

    def get_logger(self):
        if not self.logger:
            self.logger = self._create_logger()
        return self.logger

    def create_destinations(self):
        destinations = []
        if self._config.get('flush_stdout'):
            destinations.append(Stdout())
        if self._config.get('flush_graphite'):
            for graphite_address in self._config['flush_graphite'].split(','):
                graphite_address = graphite_address.strip().split(':')
                graphite_host = graphite_address.pop(0).strip()
                graphite_port = graphite_address and int(graphite_address.pop(
                )) or 2003
                destinations.append(Graphite(graphite_host, graphite_port))
        return destinations

    def create_collectors(self):
        collectors = []
        if self._config.get('collect_tcp'):
            collectors.extend(
                self._create_socket_servers(
                    self._config['collect_tcp'], socket.SOCK_STREAM))
        if self._config.get('collect_udp'):
            collectors.extend(
                self._create_socket_servers(
                    self._config['collect_udp'], socket.SOCK_DGRAM))
        if len(collectors) < 1:
            collectors.extend(
                self._create_socket_servers(
                    '127.0.0.1:8125', socket.SOCK_DGRAM))
        return collectors

    def _create_socket_servers(self, addresses, socket_type):
        collectors = []
        socket_addresses = self.get_addresses_with_unique_ports(
            addresses)
        for (host, port) in socket_addresses:
            socket_server = self._configure_socket_server(
                SocketServer(),
                host=host,
                port=port
            )
            socket_server.socket_type = socket_type
            collectors.append(socket_server)
        return collectors

    def create_server(self):
        server = Server()
        return self._configure_server(server)

    def run(self):
        with self._run_lock:
            self._register_signal_handlers()
            self._server = self.create_server()
            server_thread = Thread(target=self._server.start)
            server_thread.start()
            # If we block by joining thread, we won't receive OS signals
            # poll the status of the server thread every often instead.
            while server_thread.is_alive():
                sleep(2)
            server_thread.join()

    def shutdown(self, timeout=None):
        with self._shutdown_lock:
            self._log_debug("acquired shutdown lock")
            if self._server:
                self._log_debug("shutting down server ...")
                self._server.shutdown(timeout)
            self._server = None
            self._log("server shutdown successfully")

    def reload(self):
        with self._reload_lock:
            if not self._server:
                raise Exception("App is not running, can not reload")
            self._configure(self._args)
            logger = self._create_logger()
            self._close_logger()
            self.logger = logger
            self._configure_server(self._server)
            self._server.reload()

    def _configure_server(self, server):
        conf = self.get_config()
        logger = self.get_logger()
        destinations = self.create_destinations()
        if logger.handlers:
            server.logger = logger
        queue_processor = server.queue_processor
        if queue_processor is None:
            queue_processor = server.create_queue_processor()
        queue_processor.logger = logger
        queue_processor.flush_interval = conf['flush_interval']
        queue_processor.set_destinations(destinations)
        server.queue_processor = queue_processor
        server.set_collectors(self.create_collectors())
        return server

    def _configure_socket_server(self, collector, host=None, port=None):
        logger = self.get_logger()
        if logger.handlers:
            collector.logger = logger
        if host is not None:
            collector.host = host
        if port is not None:
            collector.port = port
        return collector

    def _configure(self, args):
        parsed_args = vars(self._parse_args(args))
        configs = self.get_default_config()
        if parsed_args['config'] is not None:
            with parsed_args['config'] as config_file:
                configs = parse_config_file(config_file)

        store_true_args = ('log_stderr', 'log_syslog', 'flush_stdout')
        for key, value in parsed_args.items():
            if key in store_true_args:
                if key not in configs or value is True:
                    configs[key] = value
                continue
            if value is not None:
                configs[key] = value

        self._config = configs

    def _parse_args(self, args):
        parser = ArgumentParser(description=self.get_description())
        parser.add_argument('-c',
                            '--config',
                            help='path to config file',
                            type=FileType('r'))
        parser.add_argument('--log-level',
                            help='logging level',
                            choices=LOG_LEVEL_NAMES)
        parser.add_argument('--log-file', help='path to log file')
        parser.add_argument('--log-stderr',
                            action='store_true',
                            help='log to stderr')
        parser.add_argument('--log-syslog',
                            action='store_true',
                            help='log to syslog')
        parser.add_argument('--syslog-socket',
                            default=default_syslog_socket(),
                            help='syslog server socket address. '
                            'socket path, or host:port for UDP'
                            )
        parser.add_argument('--flush-interval',
                            type=float,
                            help='flush interval in seconds')
        parser.add_argument('--flush-stdout',
                            action='store_true',
                            help='flush to standard output')
        parser.add_argument('--flush-graphite',
                            help='comma separated graphite addresses to flush '
                                    'stats to, each in host[:port] format',
                            default=None)
        parser.add_argument('--collect-udp',
                            help='listen on UDP addresses to collect stats')
        parser.add_argument('--collect-tcp',
                            help='listen on TCP addresses to collect stats')

        return parser.parse_args(args)

    def _register_signal_handlers(self):
        signal(SIGINT, self._handle_signal_int)
        signal(SIGTERM, self._handle_signal_term)
        #FIXME: uncomment this after fixing reload
        #signal(SIGHUP, self._handle_signal_hup)

    def _handle_signal_int(self, *args):
        self._log("received SIGINT")
        self.shutdown(3)

    def _handle_signal_term(self, *args):
        self._log("received SIGTERM")
        self.shutdown(3)

    def _handle_signal_hup(self, *args):
        self._log("received SIGHUP")
        self.reload()

    def _create_logger(self):
        log_level_name = self._config.get('log_level', 'INFO')
        if log_level_name not in LOG_LEVEL_NAMES:
            raise ValueError("invalid log level " + log_level_name)
        logger = logging.Logger('navdoon')
        logger.setLevel(getattr(logging, log_level_name))
        if self._config.get('log_stderr'):
            logger.addHandler(logging.StreamHandler(sys.stderr))
        if self._config.get('log_file'):
            logger.addHandler(logging.FileHandler(self._config['log_file']))
        if self._config.get('log_syslog'):
            syslog_address = self._config['syslog_socket'].split(':')
            if len(syslog_address) < 2:
                syslog_address = syslog_address[0].strip()
            else:
                syslog_address = tuple([syslog_address[0].strip(),
                                      int(syslog_address[1])])
            logger.addHandler(logging.handlers.SysLogHandler(syslog_address))
        return logger

    def _close_logger(self):
        if self.logger:
            handlers = []
            for handler in self.logger.handlers:
                if hasattr(handler, 'close') and callable(handler.close):
                    handler.close()
                handlers.append(handler)
            map(self.logger.removeHandler, handlers)
        self.logger = None

    @staticmethod
    def get_addresses_with_unique_ports(addresses):
        address_tuples = [tuple(address.strip().split(':')) for address in
                          addresses.split(',')]
        result = []
        ports = set()
        for address in address_tuples:
            host = address[0]
            if len(address) > 1 and address[1]:
                port_str = address[1]
                port = int(port_str)
                if port < 1 or port > 65535:
                    raise ValueError(
                        "Port {} is out of range".format(port_str))
                if port in ports:
                    raise ValueError(
                        "Port {} is already specified before".format(port_str))
            else:
                port = DEFAULT_PORT
            result.append((host, port))
            ports.add(port)
        return result


def main(args=sys.argv[1:]):
    """Entry point, setup and start the application"""
    app = App(args)
    return app.run()


if __name__ == '__main__':
    sys.exit(main())



"""
navdoon.processor
-----------------
Define queue processor, that will process the actualy Statsd requests
queued by the collectors.
"""

from time import time
from threading import Event, RLock, Thread
from navdoon.pystdlib.queue import Empty
from navdoon.utils import LoggerMixIn, DataSeries
from statsdmetrics import (Counter, Gauge, GaugeDelta, Set, Timer,
                           parse_metric_from_request)


def validate_destinations(destinations):
    for destination in destinations:
        if not hasattr(destination,
                       'flush') or not callable(destination.flush):
            raise ValueError("Invalid destination for queue processor."
                             "Destination should have a flush() method")


def validate_queue(queue_):
    if not callable(getattr(queue_, 'get', None)):
        raise ValueError("Invalid queue for queue processor."
                         "queue should have a get() method")


class QueueProcessor(LoggerMixIn):
    """Process Statsd requests queued by the collectors"""

    default_stop_process_token = None

    def __init__(self, queue_):
        validate_queue(queue_)
        super(QueueProcessor, self).__init__()
        self.log_signature = 'queue.processor '
        self.stop_process_token = self.__class__.default_stop_process_token
        self._flush_interval = 1
        self._queue = queue_
        self._should_stop_processing = Event()
        self._processing = Event()
        self._shutdown = Event()
        self._processing_lock = RLock()
        self._flush_lock = RLock()
        self._shelf = StatsShelf()
        self._destinations = []
        self._last_flush_timestamp = None

    @property
    def queue(self):
        return self._queue

    @queue.setter
    def queue(self, queue_):
        validate_queue(queue_)
        if self.is_processing():
            raise Exception(
                "Can not change queue processor's queue while processing")
        self._queue = queue_

    @property
    def flush_interval(self):
        return self._flush_interval

    @flush_interval.setter
    def flush_interval(self, interval):
        interval_float = float(interval)
        if interval_float <= 0:
            raise ValueError(
                "Invalid flush interval. Interval should be a positive number")
        self._flush_interval = interval

    def add_destination(self, destination):
        validate_destinations([destination])
        if destination not in self._destinations:
            self._destinations.append(destination)
        return self

    def clear_destinations(self):
        self._destinations = []
        return self

    def set_destinations(self, destinations):
        validate_destinations(destinations)
        self._destinations = destinations
        return self

    def is_processing(self):
        return self._processing.is_set()

    def wait_until_processing(self, timeout=None):
        return self._processing.wait(timeout)

    def process(self):
        self._log_debug("waiting for process lock ...")
        with self._processing_lock:
            self._log_debug("process lock acquired")
            if self._last_flush_timestamp is None:
                self._last_flush_timestamp = time()
            self._log("processing the queue ...")

            queue_ = self._queue
            queue_empty_error = Empty
            process = self._process_request
            stop = self._should_stop_processing
            log_debug = self._log_debug
            log = self._log
            flush = self.flush

            self._shutdown.clear()
            self._processing.set()

            try:
                while True:
                    if stop.is_set():
                        log_debug("instructed to shutdown")
                        break
                    try:
                        data = queue_.get(timeout=1)
                        queue_has_data = True
                    except queue_empty_error:
                        queue_has_data = False

                    if time(
                    ) - self._last_flush_timestamp >= self._flush_interval:
                        flush()

                    if queue_has_data:
                        if data == self.stop_process_token:
                            log("got stop process token in queue")
                            break
                        elif data:
                            process(data)
            finally:
                log("stopped processing the queue")
                self._processing.clear()
                self._shutdown.set()

    def flush(self):
        self._log_debug("waiting for flushing lock ...")
        with self._flush_lock:
            self._log_debug("flushing lock acquired")
            now = time()
            metrics = self._get_metrics_and_clear_shelf(now)
            self._log("flushing '{}' metrics to '{}' destinations".format(
                len(metrics), len(self._destinations)))
            for destination in self._destinations:
                try:
                    call_destination_thread = Thread(target=destination.flush,
                                                     args=[metrics])
                    call_destination_thread.daemon = True
                    call_destination_thread.start()
                except Exception as exp:
                    self._log_error(
                        "error occurred while flushing to destination: "
                        "{}".format(exp))
            self._last_flush_timestamp = now

    def shutdown(self):
        self._log("shutting down ...")
        self._should_stop_processing.set()

    def wait_until_shutdown(self, timeout=None):
        return self._shutdown.wait(timeout)

    def _process_request(self, request):
        self._log_debug("processing metrics: {}".format(str(request)))
        lines = [line.strip() for line in request.split("\n")]
        stop = self._should_stop_processing
        for line in lines:
            if stop.is_set():
                break
            try:
                metric = parse_metric_from_request(line)
            except ValueError as parse_error:
                self._log_error(
                    "failed to parse statsd metrics from '{}': {}".format(
                        line, parse_error))
                continue
            self._shelf.add(metric)

    def _get_metrics_and_clear_shelf(self, timestamp):
        shelf = self._shelf
        counters = shelf.counters()
        gauges = shelf.gauges()
        sets = shelf.sets()
        timers = shelf.timers()
        shelf.clear()

        metrics = []
        for name, value in counters.items():
            metrics.append((name, value, timestamp))

        for name, value in gauges.items():
            metrics.append((name, value, timestamp))

        for name, value in sets.items():
            metrics.append((name, len(value), timestamp))

        for name, timer_stats in timers.items():
            for statistic, value in timer_stats.items():
                metrics.append(
                    (
                        "{}.{}".format(name, statistic),
                        value,
                        timestamp
                    )
                )

        return metrics


class StatsShelf(object):
    """A container that will aggregate and accumulate metrics"""

    _metric_add_methods = {Counter.__name__: '_add_counter',
                           Set.__name__: '_add_set',
                           Gauge.__name__: '_add_gauge',
                           GaugeDelta.__name__: '_add_gauge_delta',
                           Timer.__name__: '_add_timer'}

    def __init__(self):
        self._lock = RLock()
        self._counters = dict()
        self._timers = dict()
        self._sets = dict()
        self._gauges = dict()

    def add(self, metric):
        method_name = self.__class__._metric_add_methods.get(
            metric.__class__.__name__)
        if not method_name:
            raise ValueError(
                "Can not add metric to shelf. No method is defined to "
                "handle {}".format(metric.__class__.__name__))
        with self._lock:
            getattr(self, method_name)(metric)

    def counters(self):
        return self._counters.copy()

    def sets(self):
        return self._sets.copy()

    def gauges(self):
        return self._gauges.copy()

    def timers_data(self):
        return self._timers.copy()

    def timers(self):
        result = dict()
        for name, timers in self._timers.items():
            series = DataSeries(timers)
            result[name] = dict(count=series.count(), min=series.min(), max=series.max(),
                                mean=series.mean(), median=series.median())
        return result

    def clear(self):
        self._counters = dict()
        self._timers = dict()
        self._sets = dict()
        self._gauges = dict()

    def _add_counter(self, counter):
        counters = self._counters
        name = counter.name
        if name not in counters:
            counters[name] = 0
        counters[name] += counter.count / counter.sample_rate

    def _add_set(self, metric):
        self._sets.setdefault(metric.name, set()).add(metric.value)

    def _add_gauge(self, metric):
        self._gauges[metric.name] = metric.value

    def _add_gauge_delta(self, metric):
        gauges = self._gauges
        name = metric.name
        if name not in gauges:
            gauges[name] = metric.delta
        else:
            gauges[name] += metric.delta

    def _add_timer(self, metric):
        self._timers.setdefault(metric.name, list()).append(
            metric.milliseconds)


#!/usr/bin/env python
"""
navdoon
"""

from __future__ import print_function

import os

try:
    import setuptools
    from setuptools import setup
except ImportError:
    setuptools = None
    from distutils.core import setup

try:
    import distutilazy.test
    import distutilazy.clean
except ImportError:
    distutilazy = None

from navdoon import (__title__, __summary__, __version__, __author__,
                     __license__)

classifiers = [
    "Development Status :: 4 - Beta", "Intended Audience :: Developers",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: System :: Networking :: Monitoring"
]

long_description = __summary__
with open(os.path.join(os.path.dirname(__file__), "README.rst")) as fh:
    long_description = fh.read()

setup_params = dict(name=__title__,
                    packages=["navdoon", "navdoon.pystdlib", "navdoon.destination"],
                    version=__version__,
                    description=__summary__,
                    long_description=long_description,
                    author=__author__,
                    url="https://github.com/farzadghanei/navdoon",
                    license=__license__,
                    classifiers=classifiers,
                    install_requires=["statsdmetrics>=0.3"])

if setuptools:
    setup_params["keywords"] = "statsd monitoring"
    setup_params["test_suite"] = "tests"
    setup_params["zip_safe"] = False
    setup_params["extras_require"] = dict(
        dev=["distutilazy>=0.4.1", "typing>=3.5.0.1", "coverage"]
    )
    setup_params["entry_points"] = dict(
        console_scripts=["navdoon=navdoon.app:main",]
    )
else:
    setup_params["scripts"] = ['bin/navdoon']


if distutilazy:
    setup_params["cmdclass"] = dict(test=distutilazy.test.run_tests,
                                    clean_pyc=distutilazy.clean.clean_pyc,
                                    clean=distutilazy.clean.clean_all)

if __name__ == "__main__":
    setup(**setup_params)

__all__ = (setup_params, classifiers, long_description)


"""
Copyright 2016 Farzad Ghanei

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import absolute_import, division, print_function

__title__ = "navdoon"
__summary__ = "Powerful Statsd server, made easy"
__version__ = "0.1.0"
__author__ = "Farzad Ghanei"
__license__ = "Apache License, Version 2.0"
__copyright__ = "Copyright 2016 Farzad Ghanei"



"""
navdoon.utils
-------------
Commont utilities and mixin classes
"""

import socket
import platform
from os import getpid
from time import sleep
from logging import INFO, DEBUG, ERROR, WARN
from threading import Lock
from multiprocessing import cpu_count


def available_cpus():
    try:
        cpus = cpu_count()
    except Exception:
        cpus = 1
    return cpus


def os_syslog_socket():
    syslog_addresses = dict(
        linux="/dev/log",
        darwin="/var/run/syslog",
        freebsd="/var/run/log"
    )
    return syslog_addresses.get(platform.system().strip().lower(), None)


class LoggerMixIn(object):
    """A MixIn class for anything that needs to log messages"""

    def __init__(self):
        self.logger = None
        self.log_pattern = "{signature}{message}"
        self.log_signature = ''
        self._pid = getpid()

    def _log_debug(self, msg):
        self._log(msg, DEBUG)

    def _log_error(self, msg):
        self._log(msg, ERROR)

    def _log_warn(self, msg):
        self._log(msg, WARN)

    def _log(self, msg, level=INFO):
        if self.logger:
            self.logger.log(
                level,
                self.log_pattern.format(message=msg,
                                        signature=self.log_signature,
                                        pid=self._pid))


class TCPClient(LoggerMixIn):
    """A generic TCP client with reconnecting feature"""

    def __init__(self, host, port):
        super(TCPClient, self).__init__()
        self.host = host
        self.port = port
        self.max_retry = None
        self._connection_tries = 0
        self._sleep_between_retries = 0.5
        self._sock = None
        self._connection_lock = Lock()
        self._sending_lock = Lock()

    def connect(self):
        self._connection()
        self._log("connected to {}:{}".format(self.host, self.port))

    def disconnect(self):
        with self._connection_lock:
            self._log_debug("disconnecting from {}:{}".format(self.host,
                                                              self.port))
            if self._sock:
                try:
                    self._sock.shutdown(socket.SHUT_RDWR)
                    self._sock.close()
                except socket.error:
                    pass
            self._sock = None
            self._log("disconnected from {}:{}".format(self.host, self.port))

    def reconnect(self):
        self._log_debug("reconnecting to {}:{}".format(self.host, self.port))
        if self._connection_tries >= self.max_retry:
            raise IOError(
                "Reached maximum connection tries of '{}' to {}:{}".format(
                    self.max_retry, self.host, self.port))
        self.disconnect()
        self.connect()

    def reset_connection_tries(self):
        with self._connection_lock:
            self._connection_tries = 0

    def _send_with_lock(self, data_bytes):
        data_size = len(data_bytes)
        with self._sending_lock:
            while True:
                self._log_debug("sending {} bytes to {}:{} ...".format(
                    data_size, self.host, self.port))
                sock = self._connection()
                try:
                    sock.sendall(data_bytes)
                    self._log_debug("sent {} bytes to {}:{}".format(
                        data_size, self.host, self.port))
                    break
                except socket.error as err:
                    self._log_error("failed to send data to {}:{}. {}".format(
                        self.host, self.port, err))
                    self.reconnect()

    def _connection(self):
        with self._connection_lock:
            if not self._sock:
                max_retry = self.max_retry
                while True:
                    self._connection_tries += 1
                    self._log_debug("connecting to {}:{} try {}/{}".format(
                        self.host, self.port, self._connection_tries,
                        max_retry))
                    try:
                        sock = socket.socket(socket.AF_INET,
                                             socket.SOCK_STREAM)
                        sock.connect((self.host, self.port))
                        self._log_debug("connected to {}:{}".format(self.host,
                                                                    self.port))
                        self._sock = sock
                        break
                    except socket.error as err:
                        self._log_error(
                            "failed to connect to {}:{}. {}".format(
                                self.host, self.port, err))
                        sock.close()
                        if max_retry and max_retry <= self._connection_tries:
                            raise IOError(
                                "Reached maximum connection tries of "
                                "'{}' to {}:{}".format(
                                    max_retry, self.host, self.port))
                    sleep(self._sleep_between_retries * self._connection_tries)
        return self._sock


class DataSeries(object):
    def __init__(self, data):
        self._count = len(data)
        if self._count < 1:
            raise ValueError("Can not create a series from an empty data set")
        self._data = sorted(data)

    def count(self):
        return self._count

    def min(self):
        return min(self._data)

    def max(self):
        return max(self._data)

    def mean(self):
        return sum(self._data) / self._count

    def median(self):
        count = self._count
        if count == 2:
            return self.mean()
        last_index = count - 1
        middle_index = count // 2
        if middle_index < last_index and count % 2 == 0:
            return (self._data[middle_index] + self._data[middle_index + 1]) / 2
        else:
            return self._data[middle_index]


"""
navdoon.destination.graphite
----------------------------
A destination to flush metrics to Graphite
"""

from time import time
from navdoon.utils import TCPClient
from navdoon.destination.abstract import AbstractDestination


class Graphite(TCPClient, AbstractDestination):
    """Flush metrics to Graphtie over a TCP connection"""

    def __init__(self, host='localhost', port=2003):
        super(Graphite, self).__init__(host, port)

    def __del__(self):
        self.disconnect()

    @staticmethod
    def create_request_from_metrics(metrics):
        """Creates Graphite protocol lines from metrics"""
        requests = []
        for metric in metrics:
            name, value = metric[:2]
            timestamp = len(metric) > 2 and metric[2] or time()
            requests.append("{} {} {}".format(name, value, timestamp))
        return requests

    def flush(self, metrics):
        """Flush metrics to Graphite"""
        lines = self.create_request_from_metrics(metrics)
        self._send_lines(lines)

    def _send_lines(self, lines):
        num_lines = len(lines)
        data = "\n".join([line.strip() for line in lines]).encode()
        self._log_debug("flushing {} metrics to graphite on {}:{} ...".format(
            num_lines, self.host, self.port))
        self._send_with_lock(data)
        self._log("flushed {} metrics to graphite on {}:{}".format(
            num_lines, self.host, self.port))

    def __eq__(self, other):
        return self.host == other.host and self.port == other.port and \
                self.max_retry == other.max_retry
